from awesomeNations.customMethods import format_key, string_is_number
from awesomeNations.exceptions import DataError
from pprint import pprint as pp
from typing import Optional
import xmltodict
import string
import random

class NationAuth():
    """Nation authentication"""
    def __init__(self,
                 password: Optional[str] = None,
                 autologin: Optional[str] = None):
        if not any((password, autologin)):
            raise ValueError("NationAuth can't be empty, a password or autologin must be given.")
        if password and type(password) is not str:
            raise ValueError(f"password must be str, not {type(password).__name__}")
        if autologin and type(autologin) is not str:
            raise ValueError(f"autologin must be str, not {type(autologin).__name__}")
        self.crip = Criptografy()
        self.password = self.__secret__(password)
        self.autologin = self.__secret__(autologin)
        self.xpin: Optional[int] = None
    
    def __secret__(self, x: str):
        hidden_x = self.crip.encrypt(x) if x else ""
        return hidden_x

    def __show__(self, x: str):
        hidden_x = self.crip.decrypt(x) if x else ""
        return hidden_x
    
    def get(self) -> dict[str]:
        auth_headers: dict[str] = {
            "X-Password": self.__show__(self.password),
            "X-Autologin": self.__show__(self.autologin),
            "X-Pin": self.xpin if self.xpin else ""
        }
        return auth_headers

class AwesomeParser():
    def __init__(self):
        pass
    
    def parse_xml(self, data: dict[str]):
        """
        Parses XML data into a dictionary.
        """
        try:
            parsed_xml: dict = xmltodict.parse(data["data"], data["encoding"], postprocessor=self.xml_postprocessor)
            return parsed_xml
        except Exception as e:
            raise DataError("XML Data", e)

    def xml_postprocessor(self, path, key: str, value: str):
        key = format_key(key, replace_empty="_", delete_not_alpha=True)
        try:
            formatted_key: str = key
            formatted_value: str = value

            if string_is_number(formatted_value):
                formatted_value: complex = complex(formatted_value).real
                formatted_value = int(formatted_value) if formatted_value.is_integer() else formatted_value
            return formatted_key, formatted_value
        except (ValueError, TypeError):
            return key, value

class Criptografy():
    "Basic substitution criptography!"
    def __init__(self):
        self.chars = " " + string.punctuation + string.digits + string.ascii_letters
        self.chars = list(self.chars)
        self.key = self.chars.copy()
        random.shuffle(self.key)
     
     # Encrypt stuff :D
    def encrypt(self, message_to_encrypt: str) -> str:
        plain_text = message_to_encrypt
        cipher_text = ""
        for letter in plain_text:
            index = self.chars.index(letter)
            cipher_text += self.key[index]
        return cipher_text

    # Decrypt stuff :D
    def decrypt(self, message_to_decrypt: str) -> str:
        cipher_text = message_to_decrypt
        plain_text = ""
        for letter in cipher_text:
            index = self.key.index(letter)
            plain_text += self.chars[index]
        return plain_text
    
    def exhaust(self):
        self.key = self.chars.copy()
        random.shuffle(self.key)

if __name__ == "__main__":
    auth = NationAuth("Hunterx")
    print(auth.password)
    print(auth.autologin)


from awesomeNations.connection import WrapperConnection, URLManager
from awesomeNations.customMethods import join_keys, format_key
from awesomeNations.internalTools import NationAuth
from awesomeNations.exceptions import HTTPError
from pprint import pprint as pp
from datetime import datetime
from typing import Optional
from urllib3 import Timeout
from typing import Literal
from pathlib import Path
from logging import WARNING
import logging

logger = logging.getLogger("AwesomeLogger")
logging.basicConfig(level=logging.WARNING, format="[%(asctime)s] %(levelname)s: %(message)s")

wrapper = WrapperConnection()
url_manager = URLManager("https://www.nationstates.net/cgi-bin/api.cgi")

class AwesomeNations():
    """
    # 🚩 AwesomeNations 🚩

    Welcome! I'm the main class of this library and can't wait to collaborate with you!
    Feel free to explore my [GitHub repository](https://github.com/Stalot/AwesomeNations)
    and report any issues [here](https://github.com/Stalot/AwesomeNations/issues).

    # 📚 Useful References 📚
    
    If you want to make things right, I highly recommend you to **read documentation**.
    Yes, I know, it sounds as thrilling as watching paint dry... But you really should!
    Here are some helpful links for coding guidelines and references. Please note that
    these resources may change over time:

    - 📖 [NationStates API Documentation](https://www.nationstates.net/pages/api.html)  
    - ⚖️ [NationStates Script Rules for HTML site](https://forum.nationstates.net/viewtopic.php?p=16394966#p16394966)
    
    ---
    
    ### user_agent:

    > Sets a User-Agent. Whenever possible, your tool should identify itself by setting
    > the User-Agent header with relevant data.

    > - `<application name>/<version> <comments>`
    > - `ExampleScript/1.2 (by:Testlandia; usedBy:Maxtopia)`
    
    ### request_timeout:

    > Defines a timeout (in seconds) for requests.

    > - `request_timeout: tuple = (10, 5)` -> 10 seconds for connecting, 5 seconds for reading.
    > - `request_timeout: int = 10` -> 10 seconds for both.
    
    ### ratelimit_sleep:
    
    > This allows to automatically "sleep" if the API ratelimit is reached, prevents temporary
    > lockouts due to excessive requests in a short span of time.

    ### ratelimit_reset_time:
    
    > Defines the reset time (in seconds) to wait when the API ratelimit is reached.
    
    ### api_version:
    
    > This setting allows you to specify the NationStates API version your script uses.

    ### log_level:
    
    > Sets logging log level, if None is given, disables logging.
    """

    def __init__(self,
                 user_agent: str,
                 request_timeout: int | tuple = (15, 10),
                 ratelimit_sleep: bool = True,
                 ratelimit_reset_time: int = 30,
                 api_version: int = 12,
                 log_level: Optional[int] = WARNING):
        self.user_agent: str = user_agent
        self.request_timeout: int | tuple = request_timeout
        self.ratelimit_sleep: bool = ratelimit_sleep
        self.ratelimit_reset_time: int = ratelimit_reset_time
        self.api_version: int = api_version
        self.log_level: Optional[int] = log_level

        headers: dict = {
        "User-Agent": self.user_agent,
        "Cache-Control": "no-cache",
        }
        
        wrapper.headers = headers
        wrapper.request_timeout = Timeout(connect=self.request_timeout[0], read=self.request_timeout[1]) if type(self.request_timeout) is tuple else int(self.request_timeout)
        wrapper.ratelimit_sleep = self.ratelimit_sleep
        wrapper.ratelimit_reset_time = self.ratelimit_reset_time
        wrapper.api_version = self.api_version
        
        if self.log_level is None:
            logger.disabled = True
        elif type(self.log_level) is int:
            logger.level = self.log_level
        else:
            raise ValueError(f"Invalid {type(self.log_level).__name__} '{self.log_level}', log_level must be an int (to change level) or None (to disable logging)")

    def __repr__(self):
        return f"AwesomeNations(user_agent={self.user_agent}, request_timeout={self.request_timeout}, ratelimit_sleep={self.ratelimit_sleep}, ratelimit_reset_time={self.ratelimit_reset_time}, api_version={self.api_version}, log_level={self.log_level})"

    def today_is_nationstates_birthday(self) -> bool:
        "Today is 11/13?"
        today = datetime.today()
        date: str = today.strftime('%D')
        birthday: bool = False
        if '11/13' in date:
            birthday = True
        return birthday

    def get_nationstates_age(self) -> int:
        "Current year - NationStates year of creation (NationStates was created in 2002)."
        created = 2002
        today = datetime.today().year
        age = today - created
        return age

    def get_daily_data_dumps(self, filepath: str | Path = "./datadump.gz", type: Literal["nation", "region"] = "nation") -> None:
        """
        Dowloads NationStates daily data dumps.
        
        ### type: str
        
        - "nation": Dowloads the nation data dump.
        - "region": Dowloads the region data dump.
        """
        nation_url: str = "https://www.nationstates.net/pages/nations.xml.gz"
        region_url: str = "https://www.nationstates.net/pages/regions.xml.gz"

        match type:
            case "nation":
                wrapper.fetch_file(nation_url, filepath)
            case "region":
                wrapper.fetch_file(region_url, filepath)
            case _:
                raise ValueError(type)

    def get_world_shards(self, shards: str | tuple[str] | list[str], **kwargs) -> dict:
        """
        Gets one or more shards from the World API.
        """
        for kwarg in kwargs:
            kwargs[kwarg] = join_keys(kwargs[kwarg])
        params: Optional[str] = join_keys([f"{kwarg}={kwargs[kwarg]}" for kwarg in kwargs], ";") if kwargs else None
        url: str = url_manager.generate_shards_url("world", shards, params)
        response: dict = wrapper.fetch_api_data(url)
        return response

    def get_world_assembly_shards(self, shards: str | tuple[str] | list[str], **kwargs) -> dict:
        """
        Gets one or more shards from the World Assembly API.
        """
        for kwarg in kwargs:
            kwargs[kwarg] = join_keys(kwargs[kwarg])
        params: Optional[str] = join_keys([f"{kwarg}={kwargs[kwarg]}" for kwarg in kwargs], ";") if kwargs else None
        url: str = url_manager.generate_shards_url("wa",
                                                   shards,
                                                   params,
                                                   council_id=kwargs["council_id"])
        response: dict = wrapper.fetch_api_data(url)
        return response

    def get_api_latest_version(self) -> int:
        """Gets NationStates API latest version"""
        url = "https://www.nationstates.net/cgi-bin/api.cgi?a=version"
        latest_version: int = int(wrapper.fetch_raw_data(url))
        return latest_version

    class Nation:
        """
        Class dedicated to NationStates nation API.
        """
        def __init__(self,
                     nation_name: str,
                     password: str = None,
                     autologin: str = None) -> None:
            self.nation_name: str = format_key(nation_name, False, '%20') # Name is automatically parsed.
            wrapper._auth = NationAuth(password, autologin) if any((password, autologin)) else None

        def __repr__(self):
            return f"Nation(nation_name={self.nation_name})"

        def exists(self) -> bool:
            """
            Checks if nation exists.
            """
            url = url_manager.generate_shards_url("nation",
                                                  None,
                                                  None,
                                                  nation_name=self.nation_name)
            status_code: int = wrapper.connection_status_code(url)
            match status_code:
                case 200:
                    return True
                case 404:
                    return False
                case _:
                    raise HTTPError(status_code)

        # DEPRECATED METHOD
        def get_public_shards(self, shards: Optional[str | tuple[str] | list[str]] = None, **kwargs) -> dict:
            """
            # THIS METHOD IS DEPRECATED
            ## Use ```get_shards()``` instead!
            
            ***
            
            Gets one or more shards from the requested nation, returns the standard API if no shards provided.
            
            ### Standard:
            
            > A compendium of the most commonly sought information.
            
            ### Shards:
            > If you don't need most of this data, please use shards instead. Shards allow you to request
            > exactly what you want and can be used to request data not available from the Standard API!
            """
            for kwarg in kwargs:
                kwargs[kwarg] = join_keys(kwargs[kwarg])
            params: Optional[str] = join_keys([f"{kwarg}={kwargs[kwarg]}" for kwarg in kwargs], ";") if kwargs else None
            url: str = url_manager.generate_shards_url("nation",
                                                       shards,
                                                       params,
                                                       nation_name=self.nation_name)
            logger.warning("get_public_shards() is deprecated.")
            response: dict = wrapper.fetch_api_data(url)
            return response

        # Replacing get_public_shards()
        def get_shards(self, shards: Optional[str | tuple[str] | list[str]] = None, **kwargs) -> dict:
            """
            Gets one or more shards from the requested nation, returns the standard API if no shards provided.
            
            ### Standard:
            
            > A compendium of the most commonly sought information.
            
            ### Shards:
            > If you don't need most of this data, please use shards instead. Shards allow you to request
            > exactly what you want and can be used to request data not available from the Standard API!
            """
            for kwarg in kwargs:
                kwargs[kwarg] = join_keys(kwargs[kwarg])
            params: Optional[str] = join_keys([f"{kwarg}={kwargs[kwarg]}" for kwarg in kwargs], ";") if kwargs else None
            url: str = url_manager.generate_shards_url("nation",
                                                       shards,
                                                       params,
                                                       nation_name=self.nation_name)
            response: dict = wrapper.fetch_api_data(url)
            return response

    class Region: 
        """
        Class dedicated to NationStates region API.
        """
        def __init__(self, region_name: str) -> None:
            # self.pretty_name: str = prettify_string(str(region_name))
            self.region_name = format_key(region_name, False, '%20')
        
        def __repr__(self):
            return f"Region(region_name={self.region_name})"
        
        def exists(self) -> bool:
            """
            Checks if region exists.
            """
            url = url_manager.generate_shards_url("region",
                                                  None,
                                                  None,
                                                  region_name=self.region_name)
            status_code: int = wrapper.connection_status_code(url)
            match status_code:
                case 200:
                    return True
                case 404:
                    return False
                case _:
                    raise HTTPError(status_code)

        def get_shards(self, shards: Optional[str | tuple[str] | list[str]] = None, **kwargs) -> dict:
            """
            Gets one or more shards from the requested region, returns the standard API if no shards provided.
            
            ### Standard:
            
            A compendium of the most commonly sought information.
            
            ### Shards:
            If you don't need most of this data, please use shards instead. Shards allow you to request exactly what you want and can be used to request data not available from the Standard API!
            """
            for kwarg in kwargs:
                kwargs[kwarg] = join_keys(kwargs[kwarg])
            params: Optional[str] = join_keys([f"{kwarg}={kwargs[kwarg]}" for kwarg in kwargs], ";") if kwargs else None
            url: str = url_manager.generate_shards_url("region",
                                                       shards,
                                                       params,
                                                       region_name=self.region_name)
            response: dict = wrapper.fetch_api_data(url)
            return response

if __name__ == "__main__":
    api = AwesomeNations("AwesomeNations/Test", log_level=0)
    print(api)
    print(api.Nation("testlandia").get_public_shards("name"))
    print(api.Region("the pacific"))


from collections.abc import Iterable
from typing import Any, Optional
import time

def string_is_number(string: str) -> bool:
    """
    Checks if string is a number or not.
    """
    if type(string) != str:
        raise TypeError(f"{type(string).__name__} {string} is not a string.")
    try:
        complex(string)
        return True
    except ValueError:
        return False

def format_key(string: str = None, uppercase: bool = False, replace_empty: str = None, delete_not_alpha: bool = False) -> str:
    """
    Formats a string.
    """
    formatted_string = None
    if string:
        if delete_not_alpha:
            for char in string:
                if not char.isalpha() and not char.isspace():
                    string = string.replace(char, '')

        if not uppercase:
            string = string.lower()
        else:
            string = string.upper()

        if replace_empty:
            string = string.replace(' ', replace_empty)

        formatted_string = string
    return formatted_string

def join_keys(keys: list[str] | tuple[str] = ['hello', 'world'], separator: str = '+') -> str:
    """
    Joins multiple string keys in a single string.
    """
    result = keys
    if isinstance(keys, Iterable) and type(keys) is not str:
        string_keys = (str(key) for key in keys)
        result = separator.join(string_keys)
    return str(result)

def generate_epoch_timestamp() -> int:
    timestamp: int = int(time.time())
    return timestamp

def get_index(object: tuple | list, index: int, default: Optional[Any] = None) -> Any | None:
    """
    Safely retrieves an element from a tuple or list by index.

    - **object** (tuple | list): The tuple or list from which to retrieve the element.
    - **index** (int): The index of the element to retrieve.
    - **default** (Any, optional): The value to return if the index is out of range. Defaults to None.

    **Returns**: The element at the specified index, or the default value if the index is out of range.
    """
    try:
        return object[index]
    except IndexError:
        return default

if __name__ == "__main__":
    ...


from awesomeNations.customMethods import join_keys, string_is_number
from awesomeNations.exceptions import HTTPError, DataError
from awesomeNations.internalTools import AwesomeParser
from awesomeNations.internalTools import NationAuth
from typing import Optional, Literal, Any
from urllib3 import BaseHTTPResponse
from pprint import pprint as pp
from pathlib import Path
import urllib3
import logging
import time

logger = logging.getLogger("AwesomeLogger")

parser = AwesomeParser()

class WrapperConnection():
    def __init__(self,
                 headers: dict = None,
                 ratelimit_sleep: bool = True,
                 ratelimit_reset_time: int = 30,
                 api_version: int = 12,
                 ):
        self.headers: dict = headers
        self.request_timeout: int | tuple = 10
        self.ratelimit_sleep: bool = ratelimit_sleep
        self.ratelimit_reset_time: int = ratelimit_reset_time
        self.ratelimit_remaining: int = None
        self.ratelimit_requests_seen: int = None
        self.api_version: int = api_version
        
        self._pool_manager = urllib3.PoolManager(4,
                                                self.headers,
                                                retries=False)
        self.last_request_headers: dict = {}
        self._auth: Optional[NationAuth] = None

    def fetch_api_data(self,
                       url: str = 'https://www.nationstates.net/',
                       query_parameters: Optional[dict] = None) -> dict:
        """
        This fetches API data and automatically parses it: (xml response -> python dictionary)
        """
        url = url.format(v=self.api_version)
        logger.debug(f"Fetching API data: {url}")
        
        # Updates headers X-Password, X-Autologin and X-Pin in the next request
        # for actions that need authentication (Like private shards).
        if self._auth:
            self.headers.update(self._auth.get())

        response = self._pool_manager.request("GET", url, headers=self.headers, fields=query_parameters, timeout=self.request_timeout)

        if response.status != 200:
            raise HTTPError(response.status)
        
        self.last_request_headers.update(response.headers)
        x_pin_header: int | None = response.headers.get("X-Pin")
        
        # Updates self._auth X-Pin if necessary (for quick sucessive requests):
        if self._auth and x_pin_header:
            if self._auth.xpin != x_pin_header:
                self._auth.xpin = x_pin_header

        self.update_ratelimit_status(response.headers)

        parsed_response = parser.parse_xml(self.decode_response_data(response))
        return parsed_response

    def fetch_raw_data(self,
                       url: str) -> str:
        logger.debug(f"Fetching raw data: {url}")
        
        response = self._pool_manager.request("GET", url)
        
        if response.status != 200:
            raise HTTPError(response.status)
        
        self.update_ratelimit_status(response.headers)
        
        return self.decode_response_data(response)["data"].strip()

    def fetch_file(self,
                   url: str,
                   filepath: str | Path) -> None:
        "Dowloads a file"
        
        logger.debug(f"Dowloading Daily Data Dump: {url}")
        
        if not Path(filepath).suffix:
            raise ValueError(f"{filepath}: This path needs a suffix dude!")
        with self._pool_manager.request("GET", url, preload_content=False) as file_response, open(filepath, "wb") as file_out:
            for chunk in file_response.stream(10**4, True):
                file_out.write(chunk)
        
        logger.debug(f"Daily Data Dump located in: {filepath}")

    def connection_status_code(self, url: str = 'https://www.nationstates.net/') -> int:
        url = url.format(v=self.api_version)
        
        logger.debug(f"Testing connection status code of: {url}")
        
        response = self._pool_manager.request("GET", url, headers=self.headers, timeout=20)
        
        self.last_request_headers.update(response.headers)
        self.update_ratelimit_status(response.headers)
        
        logger.debug(f"{url} status code is: {response.status}")

        return response.status
   
    def check_api_ratelimit(self) -> None:
        """
        Checks the NationStates API ratelimit and hibernates if the request limit was reached.
        """
        if self.ratelimit_sleep:
            if self.ratelimit_remaining != None and self.ratelimit_remaining < 1:
                    logger.warning(f"API ratelimit reached, your code will be paused for: {self.ratelimit_reset_time} seconds.")
                    time.sleep(self.ratelimit_reset_time + 1)
                    logger.info("Hibernation finished")

    def update_ratelimit_status(self, response_headers: dict) -> None:
        self.ratelimit_remaining = self.get_header(response_headers, "Ratelimit-remaining")
        self.ratelimit_requests_seen = self.get_header(response_headers, "X-ratelimit-requests-seen")
        
        logger.info(f"Ratelimit remaining: {self.ratelimit_remaining}")
        
        self.check_api_ratelimit()

    def decode_response_data(self, response: BaseHTTPResponse) -> dict[str] | None:
        encodings: tuple[str] = ("UTF-8", "LATIN-1")
        tries: int = 0
        for enc in encodings:
            try:
                data = {
                    "encoding": enc,
                    "data": response.data.decode(enc)
                }
                return data
            except Exception as decoding_error:
                logger.warning(F"Failed to decode response using {enc}")
                tries += 1
                if tries >= len(encodings):
                    raise DataError("API Response", "Decoding error.")

    def get_header(self, headers: dict, key: str, default = None) -> int | None:
        output_value: Any | None = default
        key_value: str | None = headers.get(key)
        if key_value:
            output_value = key_value
            if string_is_number(key_value):
                output_value = int(key_value)
        return output_value

class URLManager():
    def __init__(self, api_base_url: str):
        self.api_base_url = api_base_url
    
    def generate_shards_url(self,
                    modifier: Literal["nation", "region", "world", "wa"],
                    shards: Optional[str | tuple[str]] = None,
                    params: Optional[str | tuple[str]] = None,
                    **kwargs) -> str:
        """
        Generates urls for shards, returns the standard API structure if no shards provided (if supported).
        """
        querystring: str = None
        match modifier:
            case "nation":
                querystring = f"nation={kwargs["nation_name"]}&q="
                if not shards:
                    querystring = querystring.replace("&q=", "")
            case "region":
                querystring = f"region={kwargs["region_name"]}&q="
                if not shards:
                    querystring = querystring.replace("&q=", "")
            case "world":
                querystring = "q="
                if not shards:
                    raise ValueError(f"Shards cannot be None, World API modifier needs shards!")
            case "wa":
                querystring = f"wa={kwargs["council_id"]}&q="
                if not shards:
                    raise ValueError(f"Shards cannot be None, World Assembly API modifier needs shards!")
            case _:
                raise ValueError(f"{modifier} is invalid. Modifier must be nation, region, world or wa.")

        shards_query: str = shards
        shards_params: str = params
        if shards:
            if type(shards) != str:
                shards_query = join_keys(shards)
            querystring += shards_query
        if params:
            if type(params) != str:
                shards_params: str = join_keys(params, ";")
            querystring += ";" +  shards_params
        full_url: str = self.api_base_url + "?" + querystring + "&v={v}"
        return full_url

if __name__ == "__main__":
    headers = {"User-Agent": "AwesomeNations urllib3 test (by: Orlys; usdBy: Orlys)"}
    wrapper = WrapperConnection(headers)
    url_manager = URLManager("https://www.nationstates.net/cgi-bin/api.cgi")
    
    data = wrapper.fetch_api_data("https://www.nationstates.net/cgi-bin/api.cgi?nation=testlandia&q=capital")
    pp(data)